/***************************************************************************
* Copyright (C) 2011 by Adelle Hartley, T/A Akemi                           *
*                                                                           *
* Permission is hereby granted, free of charge, to any person obtaining a   *
* copy of this software and associated documentation files (the "Software"),*
* to deal in the Software without restriction, including without limitation *
* the rights to use, copy, modify, merge, publish, distribute, sublicense,  *
* and/or sell copies of the Software, and to permit persons to whom the     *
* Software is furnished to do so, subject to the following conditions:      *
*                                                                           *
* The above copyright notice and this permission notice shall be included   *
* in all copies or substantial portions of the Software.                    *
*                                                                           *
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS   *
* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF                *
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN *
* NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,  *
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR     *
* OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE *
* USE OR OTHER DEALINGS IN THE SOFTWARE.                                    *
*                                                                           *
 ***************************************************************************/

#include <fstream>
#include <ios>
#include <iostream>
#include <string>
#include <cstring>
#include <list>
#include <stdlib.h>

#define NUM_SECONDS_PER_PEAK 0.05f

#define OK 0
#define NOT_OK 1
#define RESULT int
#define PROCEED_IF_OK(res) if (res) return res;
#define RETURN_IF_OK(res) if (OK==res) return res;

std::string lastErrorMessage;

int Error(std::string message)
{
  lastErrorMessage = message;
  return NOT_OK;
}

int Verify(const char* got, const char* expected, size_t size, const char* description)
{
  RETURN_IF_OK(memcmp(got, expected, size))
  std::string message(description);
  message.append(" not as expected.");
  return Error(message);
}

int Verify(unsigned int got, unsigned int expected, size_t ignored, const char* description)
{
  if (expected==got)
  {
    return OK;
  }

  std::string message(description);
  message.append(" not as expected.");
  return Error(message);
}

#define VERIFY(got,expected,size,description) { RESULT res = Verify(got,expected,size,description); PROCEED_IF_OK(res) }

// Wav file format description gleaned from https://ccrma.stanford.edu/courses/422/projects/WaveFormat/

struct WavRiffHeader
{
  // All RIFF files start with a RIFF header that specifies the type of data, in this case "WAVE".
  char ChunkID[4]; // Should be "RIFF"
  unsigned int ChunkSize;
  char Format[4]; // Should be "WAVE"
};

struct WavFormatHeader
{
  char ChunkID[4]; // Should be "FMT"
  unsigned int ChunkSize;
  unsigned short AudioFormat;
  unsigned short NumChannels;
  unsigned int SampleRate;
  unsigned int ByteRate;
  unsigned short BlockAlign;
  unsigned short BitsPerSample;
};

struct WavDataHeader
{
  char ChunkID[4];
  unsigned int ChunkSize;
};

struct WavFileHeader
{
  WavRiffHeader RiffHeader;
  WavFormatHeader FormatHeader;
  WavDataHeader DataHeader;
};

int ReadWavFileHeader(std::istream& stream, WavFileHeader& r)
{
  memset(&r, 0, sizeof(r));  
  stream.read((char*)&r, sizeof(r));

  VERIFY(r.RiffHeader.ChunkID, "RIFF", 4, "RIFF header ChunkID")
  VERIFY(r.RiffHeader.Format, "WAVE", 4, "Riff header Format")
  VERIFY(r.FormatHeader.ChunkID, "fmt", 3, "fmt header ChunkID")
  VERIFY(r.DataHeader.ChunkID, "data", 4, "data header ChunkID")
  VERIFY(r.FormatHeader.ChunkSize, 16, 4, "fmt header chunk size")
  VERIFY(r.FormatHeader.AudioFormat, 1, 2, "Audio Format") // 1 = PCM

  return OK;
}

struct Sample
{
  char leftLo;
  char leftHi;
  char rightLo;
  char rightHi;
};

class Peak;
class PeakIterator;

class Peak
{
  friend class PeakIterator;

public:
  signed char negative;
  signed char positive;
  signed char amplitude;
  unsigned int numSamplesPerPeak;
  unsigned int numSamples;
  unsigned int numBytesPerSample;
  signed char* samples; // Allocated by caller, freed in destructor.
  unsigned int bufferSize;

public:
  Peak(unsigned int initNumSamplesPerPeak)
     : negative(0), positive(0), numSamplesPerPeak(initNumSamplesPerPeak), samples(NULL)
  {}
public:
  ~Peak()
  {
    if (samples)
    {
      free(samples);
    }
  }

  int Calc()
  {
    unsigned int numSamplesToProcess = numSamples;
    signed char* i = samples;
    negative = 127;
    positive = -128;

    while (numSamplesToProcess)
    {
      Sample& sample = *(Sample*)i;
      negative = (sample.leftHi<negative)?sample.leftHi:negative;
      negative = (sample.rightHi<negative)?sample.rightHi:negative;
      positive = (sample.leftHi>positive)?sample.leftHi:positive;
      positive = (sample.rightHi>positive)?sample.rightHi:positive;

      --numSamplesToProcess;
      i+=numBytesPerSample;
    }

    amplitude = abs(negative);
    amplitude = (positive>amplitude)?positive:amplitude;

    return OK;
  }

  unsigned int numSampleBytes()
  {
    unsigned int r = bufferSize;
    if (numSamples<numSamplesPerPeak)
    {
      r = numSamples*numBytesPerSample;
    }
    return r;
  }

};

class Peaks
{
public:
  Peaks(std::istream& stream, WavFormatHeader wavFormatHeader
             , unsigned int numSamplesPerPeak, unsigned int remainingBytes)
             : m_stream(stream), m_wavFormatHeader(wavFormatHeader)
             , m_remainingBytes(remainingBytes) , m_head(numSamplesPerPeak)
  {
    unsigned int numChannels = wavFormatHeader.NumChannels;
    unsigned int bitsPerSample = wavFormatHeader.BitsPerSample;
    m_head.numBytesPerSample = numChannels * (bitsPerSample / 8);
    m_head.bufferSize = m_head.numBytesPerSample*numSamplesPerPeak;
    m_head.samples = (signed char*)malloc(m_head.bufferSize);
    ++(*this);
  }

private:
  std::istream& m_stream;
  WavFormatHeader m_wavFormatHeader;
  unsigned int m_remainingBytes;
  unsigned int m_bufferSize;
  Peak m_head;

public:
  Peaks& operator++()
  { // preincrement
    unsigned int numBytes = (m_remainingBytes<m_head.bufferSize)?m_remainingBytes:m_head.bufferSize;
    m_head.numSamples = (m_remainingBytes<m_head.bufferSize)?numBytes/m_head.numBytesPerSample:m_head.numSamplesPerPeak;
    m_remainingBytes-=numBytes;
    m_stream.read((char*)m_head.samples, numBytes);
    m_head.Calc();

    return *this;
  }

  inline Peak& head()
  {
    return m_head;
  }

  inline Peak operator*()
  {
    return m_head;
  }

  inline bool empty()
  {
    return 0==m_remainingBytes;
  }

  void ignore(unsigned int numPeaksToIgnore)
  {
    m_stream.ignore(m_head.numBytesPerSample*m_head.numSamplesPerPeak*(numPeaksToIgnore-1));
    ++(*this);
  }

};

RESULT WriteIndividualTracks(char* inputFilename, std::list<unsigned int>& cutPoints)
{
  WavFileHeader wavFileHeader;
  std::ifstream fin(inputFilename, std::ios::binary);
  PROCEED_IF_OK(ReadWavFileHeader(fin, wavFileHeader));

  unsigned int numChannels = wavFileHeader.FormatHeader.NumChannels;
  unsigned int bitsPerSample = wavFileHeader.FormatHeader.BitsPerSample;
  unsigned int numBytesPerSample = numChannels * (bitsPerSample / 8);
  unsigned int sampleRate = wavFileHeader.FormatHeader.SampleRate/numChannels;

  unsigned int numSamplesPerPeak = NUM_SECONDS_PER_PEAK*sampleRate;
  Peaks peaks(fin, wavFileHeader.FormatHeader, numSamplesPerPeak, wavFileHeader.DataHeader.ChunkSize);

  int index = 0;
  std::list<unsigned int>::const_iterator iTrackBegin = cutPoints.begin();
  std::list<unsigned int>::const_iterator iTrackEnd = iTrackBegin;
  iTrackEnd++;
  peaks.ignore(*iTrackBegin);
  for (std::list<unsigned int>::const_iterator i=iTrackEnd; i!=cutPoints.end(); ++i)
  {
    index++;
    iTrackEnd = i;
    std::string outputFilename(inputFilename);
    outputFilename.append(" [");
    char strIndex[10];
    sprintf(strIndex, "%d", index);
    outputFilename.append(strIndex);
    outputFilename.append("].wav");
    std::cout << "Writing " << outputFilename << '\n';

    std::ofstream fout(outputFilename.c_str(), std::ofstream::binary);
    unsigned int numPeaks = *iTrackEnd-*iTrackBegin;
    unsigned int numSamples = numPeaks*numSamplesPerPeak;
    wavFileHeader.DataHeader.ChunkSize = numSamples*numBytesPerSample;
    fout.write((char*)&wavFileHeader, sizeof(wavFileHeader));
    for (unsigned int j=0; j<numPeaks; j++)
    {
      fout.write((char*)peaks.head().samples, peaks.head().numSampleBytes());
      ++peaks;
    }

    fout.close();
    iTrackBegin = iTrackEnd;
  }

  return OK;
}

RESULT ProcessWavFile(char* filename)
{
  std::cout << "Reading " << filename << '\n';

  WavFileHeader wavFileHeader;
  std::ifstream fin(filename, std::ios::binary);
  PROCEED_IF_OK(ReadWavFileHeader(fin, wavFileHeader));

  unsigned int numChannels = wavFileHeader.FormatHeader.NumChannels;
  unsigned int sampleRate = wavFileHeader.FormatHeader.SampleRate/numChannels;
  unsigned int bitsPerSample = wavFileHeader.FormatHeader.BitsPerSample;
  unsigned int numBytesPerSample = numChannels * (bitsPerSample / 8);
  unsigned int blockAlign = wavFileHeader.FormatHeader.BlockAlign;
  VERIFY(blockAlign, numBytesPerSample, 2, "BlockAlign")
  VERIFY(bitsPerSample, 16, 2, "Bits per sample")

  // TODO: Handle other combinations of bits per sample & num channels.
  VERIFY(numBytesPerSample, 4, 2, "Bytes per sample")

  unsigned int numSamplesPerPeak = NUM_SECONDS_PER_PEAK*sampleRate;
  Peaks peaks(fin, wavFileHeader.FormatHeader, numSamplesPerPeak, wavFileHeader.DataHeader.ChunkSize);

  unsigned int runLength = 0;
  unsigned int numCounterExamples = 0;
  unsigned int threshold = 6;
  float numSecondsSilence = 1.2;
  float numSecondsNonSilence = 2.0;
  unsigned int minExamples = numSecondsSilence*sampleRate/numSamplesPerPeak;
  float maxSecondsFinalSilence = 6;
  unsigned int maxPeaksFinalSilence = maxSecondsFinalSilence*sampleRate/numSamplesPerPeak;
  unsigned int maxCounterExamples = numSecondsNonSilence*sampleRate/numSamplesPerPeak;
  std::list<unsigned int> cutPoints;
  unsigned int lastSilence = 0;
  float minSecondsPerTrack = 20;
  unsigned int minPeaksPerTrack = minSecondsPerTrack*sampleRate/numSamplesPerPeak;
  unsigned int lastCutPoint = 0;
  float numSecondsSilenceToKeepAtBeginningOfTrack = 0.2;
  unsigned int numPeaksSilenceToKeepAtBeginningOfTrack = sampleRate*numSecondsSilenceToKeepAtBeginningOfTrack/numSamplesPerPeak;
  float numSecondsNeedlePlop = 0.1;
  unsigned int numPeaksNeedlePlop = numSecondsNeedlePlop*sampleRate/numSamplesPerPeak;
  unsigned int thresholdNeedlePlop =  50;
  unsigned int numPeaksHeardNeedlePlop = 0;
  unsigned int iLastNeedlePlop = 0;

  unsigned int numSamples = wavFileHeader.DataHeader.ChunkSize / blockAlign;
  unsigned int i = 0;

  while (!peaks.empty())
  {
    signed char amplitude = peaks.head().amplitude;
    if (amplitude>=threshold)
    {
      if (runLength)
      {
        numCounterExamples++;
        if (amplitude>=thresholdNeedlePlop)
        {
          numPeaksHeardNeedlePlop++;
          if (cutPoints.empty() && numPeaksHeardNeedlePlop>=numPeaksNeedlePlop)
          {
            iLastNeedlePlop = i;
          }
        }
      }
      if (numCounterExamples>maxCounterExamples)
      {
        if (cutPoints.empty() || runLength>=minExamples)
        {
          if (cutPoints.empty() || (i-lastCutPoint)>=minPeaksPerTrack)
          {
            lastCutPoint = i-numCounterExamples-numPeaksSilenceToKeepAtBeginningOfTrack;
            cutPoints.push_back(lastCutPoint);
          }
        }
        runLength = 0;
      }
    }
    else
    {
      runLength++;
      if (runLength>=minExamples)
      {
        if (runLength==minExamples)
        {
          numCounterExamples = 0;
        }
        if (runLength<=maxPeaksFinalSilence)
        {
          lastSilence = i; // cut a little after where the last silence was first detected, to catch any "fade-out".
        }
        if (numPeaksHeardNeedlePlop>=numPeaksNeedlePlop && cutPoints.empty())
        {
          if (i-runLength<iLastNeedlePlop+numPeaksNeedlePlop*4)
          { // Arbitrary multiple of 4 for "typical" needle plop (numSamplesNeedlePlop is a minimum).
            lastCutPoint = iLastNeedlePlop+numPeaksNeedlePlop*4;
            if (lastCutPoint>i)
            {
              i = lastCutPoint;
            }
          }
          else
          {
            lastCutPoint = i-runLength;
          }
          cutPoints.push_back(lastCutPoint);
        }
      }
    } 

    ++peaks;
    ++i;
  }

  if (runLength<minExamples)
  {
    cutPoints.push_back(numSamples-1);
  }
  else
  {
    if ((lastSilence - lastCutPoint)>minPeaksPerTrack)
    {
      cutPoints.push_back(lastSilence);
    }
  }

  fin.close();

  return WriteIndividualTracks(filename, cutPoints);
}

int main(int argc, char* argv[])
{
  char* filename = NULL;

  // Parse command line arguments.
  for (int i=1; i<argc; i++) // Skip 0th argument, since that's just our name.
  {
    if (argv[i])
    {
      if (*argv[i])
      {
        if ('-'==*argv[i])
        {
          
        }
        else
        {
          filename = argv[i];
        }
      }
    }
  }

  RESULT res = ProcessWavFile(filename);
  RETURN_IF_OK(res);

  // An error occurred.
  // Add some detail to the error message.
  std::string message("[lpslice] Error processing ");
  message.append(filename);
  if (!lastErrorMessage.empty())
  {
    message.append(": ");
    message.append(lastErrorMessage);
  }
  lastErrorMessage = message;

  // Send the error message to stdout.
  if (!lastErrorMessage.empty())
  {
    std::cout << lastErrorMessage << '\n';
  }

}

